/*
 *  Copyright (C) 2007 mark307
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <wchar.h>
#include <tchar.h>
#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>
#include <resource.h>

#ifndef WM_CLOSETAB
#define WM_CLOSETAB WM_APP + 150
#endif

#define PLUGIN_NAME _T("Undo latest closed page plugin")
#define PREF_KEY_MAX _T("kmeleon.plugins.undoclosed.max")
#define PREF_KEY_OPEN _T("kmeleon.plugins.undoclosed.open")

#define MAX_HISTORY 20 // 10~30 is good.
#define MAX_LABEL 40

#define OPEN_LAYER 32
#define OPEN_BACKGROUNDLAYER 33

long DoMessage(LPCTSTR, const char *, const char *, long, long);
LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
void DoLoad();
void DoConfig(HWND);
void DoCreate(HWND);
void DoClose(HWND);
void DoMenu(HMENU, LPCTSTR);
int DoAccel(LPCTSTR);
void DoQuit();
void DoUndo(HWND, int);

kmeleonPlugin kPlugin = {
	KMEL_PLUGIN_VER,
	PLUGIN_NAME,
	DoMessage
};

WNDPROC KMeleonWndProc;
INT g_id_clear = -1;
INT g_id_config = -1;
INT g_id_rewind = -1;
INT g_id_start = 0;
INT g_max = MAX_HISTORY;
INT g_open_flag;
BOOL g_is_layers = FALSE;

typedef struct menulist {
	HMENU hMenu;
	struct menulist *next;
} MenuList;
MenuList *g_MenuList;

typedef struct {
	LPTSTR title;
	LPTSTR url;
} PageList;
PageList *g_PageList;


void trace(LPCTSTR lpFormat, ...)
{
#ifndef NDEBUG
	TCHAR lpMessage[256];
	va_list ap;
	
	va_start(ap, lpFormat);
	_vsntprintf(lpMessage, 256, lpFormat, ap);
	va_end(ap);
	OutputDebugString(lpMessage);
#endif
}


long DoMessage(LPCTSTR to, LPCTSTR from, LPCTSTR subject, long data1, long data2)
{
	if (to[0] == _T('*') || lstrcmpi(to, kPlugin.dllname) == 0) {
		trace(subject);
		if (lstrcmpi(subject, _T("Load")) == 0) {
			DoLoad();
		}
		else if (lstrcmpi(subject, _T("Create")) == 0) {
			if (g_PageList)
				DoCreate((HWND) data1);
		}
		else if (lstrcmpi(subject, _T("Close")) == 0) {
			if (g_PageList)
				DoClose((HWND) data1);
		}
/*
  // Not working on 1.5a1
		else if (lstrcmpi(subject, _T("CloseTab")) == 0) {
			trace("CloseTab");
			if (g_PageList)
				DoClose((HWND) data2);
		}
		else if (lstrcmpi(subject, _T("DestroyTab")) == 0) {
			trace("DestroyTab");
			if (g_PageList)
				DoClose((HWND) data2);
		}
*/
		else if (lstrcmpi(subject, _T("DoAccel")) == 0) {
			if (g_PageList) {
				LPINT id = (LPINT) data2;
				*id = DoAccel((LPCTSTR) data1);
			}
		}
		else if (lstrcmpi(subject, _T("DoMenu")) == 0) {
			if (g_PageList)
				DoMenu((HMENU) data1, (LPCTSTR) data2);
		}
		else if (stricmp(subject, "Quit") == 0) {
			if (g_PageList)
				DoQuit();
		}
		else return 0;
		return 1;
	}
	return 0;
}


void DoLoad()
{
	kPlugin.kFuncs->GetPreference(PREF_INT, PREF_KEY_MAX, &g_max, &g_max);
	kPlugin.kFuncs->GetPreference(PREF_INT, PREF_KEY_OPEN, &g_open_flag, &g_open_flag);
	kPlugin.kFuncs->GetPreference(PREF_BOOL, _T("kmeleon.plugins.layers.load"), &g_is_layers, &g_is_layers);
	
	if (g_max < 1) g_max = 1;
	else if (g_max > 100) g_max = MAX_HISTORY;
	
	g_PageList = (PageList *) calloc(g_max, sizeof(PageList));
	
	g_id_clear = kPlugin.kFuncs->GetCommandIDs(1);
	g_id_config = kPlugin.kFuncs->GetCommandIDs(1);
	g_id_start = kPlugin.kFuncs->GetCommandIDs(g_max);
	g_id_rewind = kPlugin.kFuncs->GetCommandIDs(g_max);
}


BOOL CALLBACK
DlgProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static const struct {
		UINT flag, id;
	} flagmap[] = {
		{OPEN_NORMAL, 100, },
		{OPEN_NEW, 101, },
		{OPEN_BACKGROUND, 102, },
#ifdef OPEN_NEWTAB
		{OPEN_NEWTAB, 103, },
#endif
#ifdef OPEN_BACKGROUNDTAB
		{OPEN_BACKGROUNDTAB, 104, },
#endif
		{OPEN_LAYER, 105, },
		{OPEN_BACKGROUNDLAYER, 106, },
	};
	static const int len = sizeof(flagmap) / sizeof(flagmap[0]);
	int i, n;
	
	switch (message) {
	case WM_INITDIALOG:
		SendDlgItemMessage(hwnd, 108, UDM_SETRANGE, 0, MAKELONG(30, 1));
		SendDlgItemMessage(hwnd, 108, UDM_SETPOS, 0, MAKELONG(HIWORD(lParam), 0));
		CheckRadioButton(hwnd, flagmap[0].id, flagmap[len - 1].id, flagmap[0].id);
		for (i = 0; i < len; i ++) {
			if (flagmap[i].flag == LOWORD(lParam)) {
				CheckRadioButton(hwnd, flagmap[0].id, flagmap[len - 1].id, flagmap[i].id);
			}
		}
#ifdef OPEN_NEWTAB
		if (kPlugin.kFuncs->GetKmeleonVersion() < 0x01050000) {
#endif
			EnableWindow(GetDlgItem(hwnd, 103), FALSE);
			if (IsDlgButtonChecked(hwnd, 103)) {
				CheckRadioButton(hwnd, flagmap[0].id, flagmap[len - 1].id, flagmap[0].id);
			}
#ifdef OPEN_NEWTAB
		}
#endif
#ifdef OPEN_BACKGROUNDTAB
		if (kPlugin.kFuncs->GetKmeleonVersion() < 0x01050000) {
#endif
			EnableWindow(GetDlgItem(hwnd, 104), FALSE);
			if (IsDlgButtonChecked(hwnd, 104)) {
				CheckRadioButton(hwnd, flagmap[0].id, flagmap[len - 1].id, flagmap[0].id);
			}
#ifdef OPEN_BACKGROUNDTAB
		}
#endif
		if (! g_is_layers) {
			EnableWindow(GetDlgItem(hwnd, 105), FALSE);
			EnableWindow(GetDlgItem(hwnd, 106), FALSE);
			if (IsDlgButtonChecked(hwnd, 105) || IsDlgButtonChecked(hwnd, 106)) {
				CheckRadioButton(hwnd, flagmap[0].id, flagmap[len - 1].id, flagmap[0].id);
			}
		}
		return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDOK:
			n = SendDlgItemMessage(hwnd, 108, UDM_GETPOS, 0, 0);
			if (n < 1) n = 1;
			else if (n > 30) n = 30;
			for (i = 0; i < len; i ++) {
				if (IsDlgButtonChecked(hwnd, flagmap[i].id)) {
					break;
				}
			}
			if (i < len) {
				EndDialog(hwnd, MAKELONG(flagmap[i].flag, n));
				break;
			}
		case IDCANCEL:
			EndDialog(hwnd, 0);
			break;
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


void DoConfig(HWND hwnd)
{
	LONG max = g_max, flag = g_open_flag, n;
	
	kPlugin.kFuncs->GetPreference(PREF_INT, PREF_KEY_MAX, &max, &max);
	kPlugin.kFuncs->GetPreference(PREF_INT, PREF_KEY_OPEN, &flag, &flag);
	n = MAKELONG(flag, max);
	n = DialogBoxParam(kPlugin.hDllInstance, _T("DIALOG_CONFIG"), hwnd, DlgProc, n);
	if (n == 0) return;
	max = HIWORD(n);
	flag = LOWORD(n);
	kPlugin.kFuncs->SetPreference(PREF_INT, PREF_KEY_MAX, &max, TRUE);
	kPlugin.kFuncs->SetPreference(PREF_INT, PREF_KEY_OPEN, &flag, TRUE);
}


void DoCreate(HWND hWndParent)
{
	KMeleonWndProc = (WNDPROC) GetWindowLong(hWndParent, GWL_WNDPROC);
	SetWindowLong(hWndParent, GWL_WNDPROC, (LONG)WndProc);
}


int DoAccel(LPCTSTR param)
{
	int n;
	
	if (lstrcmpi(param, _T("Clear")) == 0) {
		return g_id_clear;
	}
	else if (lstrcmpi(param, _T("Config")) == 0) {
		return g_id_config;
	}
	else if (lstrcmpi(param, _T("Rewind")) == 0) {
		return g_id_rewind;
	}
	n = _ttoi(param);
	return (n >= 0 && n < g_max) ? g_id_start + n : -1;
}


void DoMenu(HMENU hmenu, LPCTSTR param)
{
	if (param && param[0]) {
		LPCTSTR action = param, label, sep;
		sep = _tcschr(param, _T(','));
		if (sep) {
			label = sep;
			while (_istspace(*label)) label ++;
		} else {
			label = param;
			sep = param + lstrlen(param);
		}
		int command = 0;
		if (! _tcsnicmp(action, _T("Clear"), sep - action)) {
			command = g_id_clear;
		}
		else if (! _tcsnicmp(action, _T("Config"), sep - action)) {
			command = g_id_config;
		}
		else if (! _tcsnicmp(action, _T("Rewind"), sep - action)) {
			command = g_id_rewind;
		}
		if (command && label && *label) {
			AppendMenu(hmenu, MF_STRING, command, label);
		}
		return;
	}
	MenuList *item;
	item = (MenuList *) malloc(sizeof(MenuList));
	if (! item) return;
	item->hMenu = hmenu;
	item->next = g_MenuList;
	g_MenuList = item;
}


void DoQuit()
{
	MenuList *item;
	int i;
	
	while (g_MenuList) {
		item = g_MenuList;
		g_MenuList = g_MenuList->next;
		free(item);
	}
	for (i = 0; i < g_max; i ++) {
		free(g_PageList[i].title);
		free(g_PageList[i].url);
	}
	free(g_PageList);
}


void DoClear()
{
	MenuList *m;
	INT i;
	for (m = g_MenuList; m != NULL; m = m->next) {
		for (i = 0; i < g_max; i ++) {
			DeleteMenu(m->hMenu, i + g_id_start, MF_BYCOMMAND);
		}
	}
	for (i = 0; i < g_max; i ++) {
		free(g_PageList[i].title);
		g_PageList[i].title = NULL;
		free(g_PageList[i].url);
		g_PageList[i].url = NULL;
	}
}


BOOL ignore_url(LPCTSTR url)
{
	static LPCTSTR ignores[] = {
		_T("about:"),
		_T("chrome:"),
		_T("wyciwyg:"),
		NULL,
	};
	int i;
	for (i = 0; ignores[i]; i ++) {
		if (! _tcsncmp(url, ignores[i], lstrlen(ignores[i]))) {
			return TRUE;
		}
	}
	return FALSE;
}


void update_menu()
{
	MenuList *m;
	int i;
	
	for (m = g_MenuList; m != NULL; m = m->next) {
		for (i = 0; i < g_max; i ++) {
			DeleteMenu(m->hMenu, i + g_id_start, MF_BYCOMMAND);
		}
		for (i = 0; i < g_max && g_PageList[i].title; i ++) {
			AppendMenu(m->hMenu, MF_STRING, i + g_id_start, g_PageList[i].title);
		}
	}
}


void DoClose(HWND hwnd)
{
	static const TCHAR omit[] = _T(" ...");
	kmeleonDocInfo *docinfo;
	LPCTSTR p, q;
	LPTSTR title, url;
	int i, length, size;
	
	docinfo = kPlugin.kFuncs->GetDocInfo(hwnd);
	if (! docinfo || ! docinfo->url || ! docinfo->url[0]) return;
	if (ignore_url(docinfo->url)) return;
	
	for (i = 0; i < g_max && g_PageList[i].url; i ++) {
		if (! _tcscmp(docinfo->url, g_PageList[i].url)) {
			if (i > 0) {
				title = g_PageList[i].title;
				url = g_PageList[i].url;
				for (; i > 0; i --) {
					g_PageList[i].title = g_PageList[i - 1].title;
					g_PageList[i].url = g_PageList[i - 1].url;
				}
				g_PageList[0].title = title;
				g_PageList[0].url = url;
				update_menu();
			}
			return;
		}
	}
	
	p = q = (docinfo->title && docinfo->title[0]) ? docinfo->title : docinfo->url;
	while (*p && p - q < MAX_LABEL) {
		p = CharNext(p);
	}
	length = p - q + 1;
	size = (*p) ? length + lstrlen(omit) : length;
	title = (LPTSTR) malloc(size * sizeof(TCHAR));
	if (! title) return;
	lstrcpyn(title, q, length);
	if (*p) {
		lstrcat(title, omit);
	}
	
	size = lstrlen(docinfo->url) + 1;
	url = (LPTSTR) malloc(size * sizeof(TCHAR));
	if (! url) {
		free(title);
		return;
	}
	lstrcpy(url, docinfo->url);
	
	free(g_PageList[g_max - 1].title);
	free(g_PageList[g_max - 1].url);
	for (i = g_max - 1; i > 0; i --) {
		g_PageList[i].title = g_PageList[i - 1].title;
		g_PageList[i].url = g_PageList[i - 1].url;
	}
	g_PageList[0].title = title;
	g_PageList[0].url = url;
	
	update_menu();
}


void DoUndo(HWND hwnd, int n)
{
	if (! g_PageList || n < 0 || n >= g_max || ! g_PageList[n].url) return;
	switch (g_open_flag) {
	case OPEN_NEW:
	case OPEN_BACKGROUND:
#ifdef OPEN_NEWTAB
	case OPEN_NEWTAB:
#endif
#ifdef OPEN_BACKGROUNDTAB
	case OPEN_BACKGROUNDTAB:
#endif
#if (KMEL_PLUGIN_VER) < 0x0202
		kPlugin.kFuncs->NavigateTo(g_PageList[n].url, g_open_flag, hwnd);
#else
		kPlugin.kFuncs->_NavigateTo(g_PageList[n].url, g_open_flag, hwnd);
#endif
		break;
	case OPEN_LAYER:
		kPlugin.kFuncs->SendMessage(_T("layers"), kPlugin.dllname, _T("OpenURL"), (long) g_PageList[n].url, 0);
		break;
	case OPEN_BACKGROUNDLAYER:
		kPlugin.kFuncs->SendMessage(_T("layers"), kPlugin.dllname, _T("OpenURLBg"), (long) g_PageList[n].url, 0);
		break;
	default:
#if (KMEL_PLUGIN_VER) < 0x0202
		kPlugin.kFuncs->NavigateTo(g_PageList[n].url, OPEN_NORMAL, hwnd);
#else
		kPlugin.kFuncs->_NavigateTo(g_PageList[n].url, OPEN_NORMAL, hwnd);
#endif
		break;
	}
}


BOOL get_url_prefix(LPCTSTR p, LPCTSTR *last)
{
	// RFC3986 Appendix B.
	if (! *p || *p == _T(':') || *p == _T('/') || *p == _T('?') || *p == _T('#')) return FALSE;
	p = CharNext(p);
	while (*p != _T(':')) {
		if (! *p || *p == _T('/') || *p == _T('?') || *p == _T('#')) return FALSE;
		p = CharNext(p);
	}
	p = CharNext(p);
	if (*p != _T('/')) return FALSE;
	p = CharNext(p);
	if (*p != _T('/')) return FALSE;
	do {
		p = CharNext(p);
		if (*p == _T('?') || *p == _T('#')) return FALSE;
	} while (*p && *p != _T('/'));
	*last = p;
	return TRUE;
}


void DoRewind(HWND hwnd)
{
	LPTSTR *titles, *urls;
	LPCTSTR p;
	int index, count, i, length;
	
	count = index = 0;
	if (! kPlugin.kFuncs->_GetMozillaSessionHistory (&titles, &urls, &count, &index)) return;
	if (! titles || ! urls || ! index || ! count) return;
	p = urls[index];
	if (! p) return;
	if (ignore_url(p)) return;
	if (! get_url_prefix(p, &p)) return;
	length = p - urls[index];
	i = index;
	while (i > 0) {
		i --;
		if (_tcsncmp(urls[index], urls[i], length)) {
			i ++;
			break;
		}
	}
	if (i < index) {
		kPlugin.kFuncs->GotoHistoryIndex(i);
	}
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
	WPARAM wParam, LPARAM lParam)
{
	INT id;
	
	switch (message) {
	case WM_COMMAND:
		id = LOWORD(wParam);
		if (id == g_id_clear) {
			DoClear();
		} else if (id == g_id_config) {
			DoConfig(hWnd);
		} else if (id == g_id_rewind) {
			DoRewind(hWnd);
#ifdef ID_CLOSE_TAB
		} else if (id == ID_CLOSE_TAB) {
			trace("ID_CLOSE_TAB");
			DoClose(hWnd);
#endif
		} else if (id >= g_id_start && id < g_id_start + g_max) {
			DoUndo(hWnd, id - g_id_start);
		}
		break;
	case WM_CLOSETAB:
		trace("WM_CLOSETAB");
		DoClose(hWnd);
		break;
	case WM_MENUSELECT:
		id = LOWORD(wParam);
		if (id >= g_id_start && id < g_id_start + g_max) {
			int n = id - g_id_start;
			if (g_PageList && n >= 0 && n < g_max && g_PageList[n].url) {
				kPlugin.kFuncs->SetStatusBarText(g_PageList[n].url);
				return 0;
			}
		}
		break;
	}
	return CallWindowProc(KMeleonWndProc, hWnd, message, wParam, lParam);
}


extern "C" {
	BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
	{
		switch (dwReason) {
		case DLL_PROCESS_DETACH:
			break;
		}
		return TRUE;
	}
	
	KMELEON_PLUGIN kmeleonPlugin *GetKmeleonPlugin() {
		return &kPlugin;
	}
	
	KMELEON_PLUGIN int DrawBitmap(DRAWITEMSTRUCT *dis) {
		return 14; // 14 = icon width
	}
}
